#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tools for working with view projections for 2- and 3-D rendering.

"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = ['Frustum',
           'visualAngle',
           'computeFrustum',
           'computeFrustumFOV',
           'projectFrustum',
           'projectFrustumToPlane',
           'generalizedPerspectiveProjection',
           'orthoProjectionMatrix',
           'perspectiveProjectionMatrix',
           'lookAt',
           'pointToNdc',
           'cursorToRay',
           'visible',
           'visibleBBox']

import numpy as np
from collections import namedtuple
import psychopy.tools.mathtools as mt

DEG_TO_RAD = np.pi / 360.0
VEC_FWD_AND_UP = np.array(((0., 0., -1.), (0., 1., 0.)), dtype=np.float32)


def visualAngle(size, distance, degrees=True, out=None, dtype=None):
    """Get the visual angle for an object of `size` at `distance`. Object is
    assumed to be fronto-parallel with the viewer.

    This function supports vector inputs. Values for `size` and `distance` can
    be arrays or single values. If both inputs are arrays, they must have the
    same size.

    Parameters
    ----------
    size : float or array_like
        Size of the object in meters.
    distance : float or array_like
        Distance to the object in meters.
    degrees : bool
        Return result in degrees, if `False` result will be in radians.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    float
        Visual angle.

    Examples
    --------
    Calculating the visual angle (vertical FOV) of a monitor screen::

        monDist = 0.5  # monitor distance, 50cm
        monHeight = 0.45  # monitor height, 45cm

        vertFOV = visualAngle(monHeight, monDist)

    Compute visual angle at multiple distances for objects with the same size::

        va = visualAngle(0.20, [1.0, 2.0, 3.0])  # returns
        # [11.42118627  5.72481045  3.81830487]

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    size, distance = np.atleast_1d(size, distance)

    if out is not None:
        out[:] = 2 * np.arctan(size / (2 * distance), dtype=dtype)
        if degrees:
            out[:] = np.degrees(out, dtype=dtype)
        toReturn = out
    else:
        toReturn = 2 * np.arctan(size / (2 * distance), dtype=dtype)
        if degrees:
            toReturn[:] = np.degrees(toReturn, dtype=dtype)

    return toReturn


# convenient named tuple for storing frustum parameters
Frustum = namedtuple(
    'Frustum',
    ['left', 'right', 'bottom', 'top', 'nearVal', 'farVal'])


def computeFrustum(scrWidth,
                   scrAspect,
                   scrDist,
                   convergeOffset=0.0,
                   eyeOffset=0.0,
                   nearClip=0.01,
                   farClip=100.0,
                   dtype=None):
    """Calculate frustum parameters. If an eye offset is provided, an asymmetric
    frustum is returned which can be used for stereoscopic rendering.

    Parameters
    ----------
    scrWidth : float
        The display's width in meters.
    scrAspect : float
        Aspect ratio of the display (width / height).
    scrDist : float
        Distance to the screen from the view in meters. Measured from the center
        of their eyes.
    convergeOffset : float
        Offset of the convergence plane from the screen. Objects falling on this
        plane will have zero disparity. For best results, the convergence plane
        should be set to the same distance as the screen (0.0 by default).
    eyeOffset : float
        Half the inter-ocular separation (i.e. the horizontal distance between
        the nose and center of the pupil) in meters. If eyeOffset is 0.0, a
        symmetric frustum is returned.
    nearClip : float
        Distance to the near clipping plane in meters from the viewer. Should be
        at least less than `scrDist`.
    farClip : float
        Distance to the far clipping plane from the viewer in meters. Must be
        >nearClip.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        Array of frustum parameters. Can be directly passed to
        glFrustum (e.g. glFrustum(*f)).

    Notes
    -----

    * The view point must be transformed for objects to appear correctly.
      Offsets in the X-direction must be applied +/- eyeOffset to account for
      inter-ocular separation. A transformation in the Z-direction must be
      applied to account for screen distance. These offsets MUST be applied to
      the GL_MODELVIEW matrix, not the GL_PROJECTION matrix! Doing so may break
      lighting calculations.

    Examples
    --------

    Creating a frustum and setting a window's projection matrix::

        scrWidth = 0.5  # screen width in meters
        scrAspect = win.size[0] / win.size[1]
        scrDist = win.scrDistCM * 100.0  # monitor setting, can be anything
        frustum = viewtools.computeFrustum(scrWidth, scrAspect, scrDist)

    Accessing frustum parameters::

        left, right, bottom, top, nearVal, farVal = frustum
        # ... or ...
        left = frustum.left

    Off-axis frustums for stereo rendering::

        # compute view matrix for each eye, these value usually don't change
        eyeOffset = (-0.035, 0.035)  # +/- IOD / 2.0
        scrDist = 0.50  # 50cm
        scrWidth = 0.53  # 53cm
        scrAspect = 1.778
        leftFrustum = viewtools.computeFrustum(
            scrWidth, scrAspect, scrDist, eyeOffset[0])
        rightFrustum = viewtools.computeFrustum(
            scrWidth, scrAspect, scrDist, eyeOffset[1])
        # make sure your view matrix accounts for the screen distance and eye
        # offsets!

    Using computed view frustums with a window::

        win.projectionMatrix = viewtools.perspectiveProjectionMatrix(*frustum)
        # generate a view matrix looking ahead with correct viewing distance,
        # origin is at the center of the screen. Assumes eye is centered with
        # the screen.
        eyePos = [0.0, 0.0, scrDist]
        screenPos = [0.0, 0.0, 0.0]  # look at screen center
        eyeUp = [0.0, 1.0, 0.0]
        win.viewMatrix = viewtools.lookAt(eyePos, screenPos, eyeUp)
        win.applyViewTransform()  # call before drawing

    """
    # mdc - uses display size instead of the horizontal FOV gluPerspective needs
    d = scrWidth / 2.0
    ratio = nearClip / float((convergeOffset + scrDist))

    right = (d - eyeOffset) * ratio
    left = (d + eyeOffset) * -ratio
    top = d / float(scrAspect) * ratio
    bottom = -top

    return np.asarray((left, right, bottom, top, nearClip, farClip),
                      dtype=dtype)


def computeFrustumFOV(scrFOV,
                      scrAspect,
                      scrDist,
                      convergeOffset=0.0,
                      eyeOffset=0.0,
                      nearClip=0.01,
                      farClip=100.0,
                      dtype=None):
    """Compute a frustum for a given field-of-view (FOV).

    Similar to `computeFrustum`, but computes a frustum based on FOV rather than
    screen dimensions.

    Parameters
    ----------
    scrFOV : float
        Vertical FOV in degrees (fovY).
    scrAspect : float
        Aspect between the horizontal and vertical FOV (ie. fovX / fovY).
    scrDist : float
        Distance to the screen from the view in meters. Measured from the center
        of the viewer's eye(s).
    convergeOffset : float
        Offset of the convergence plane from the screen. Objects falling on this
        plane will have zero disparity. For best results, the convergence plane
        should be set to the same distance as the screen (0.0 by default).
    eyeOffset : float
        Half the inter-ocular separation (i.e. the horizontal distance between
        the nose and center of the pupil) in meters. If eyeOffset is 0.0, a
        symmetric frustum is returned.
    nearClip : float
        Distance to the near clipping plane in meters from the viewer. Should be
        at least less than `scrDist`. Never should be 0.
    farClip : float
        Distance to the far clipping plane from the viewer in meters. Must be
        >nearClip.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Examples
    --------
    Equivalent to `gluPerspective`::

        frustum =  computeFrustumFOV(45.0, 1.0, 0.5)
        projectionMatrix = perspectiveProjectionMatrix(*frustum)
    """
    d = np.tan(scrFOV * DEG_TO_RAD)
    ratio = nearClip / float((convergeOffset + scrDist))

    right = (d - eyeOffset) * ratio
    left = (d + eyeOffset) * -ratio
    top = d / float(scrAspect) * ratio
    bottom = -top

    return np.asarray((left, right, bottom, top, nearClip, farClip),
                      dtype=dtype)


def projectFrustum(frustum, dist, dtype=None):
    """Project a frustum on a fronto-parallel plane and get the width and height
    of the required drawing area.

    This function can be used to determine the size of the drawing area required
    for a given frustum on a screen. This is useful for cases where the observer
    is viewing the screen through a physical aperture that limits the FOV to a
    sub-region of the display. You must convert the size in meters to units of
    your screen and apply any offsets.

    Parameters
    ----------
    frustum : array_like
        Frustum parameters (left, right, bottom, top, near, far), you can
        exclude `far` since it is not used in this calculation. However, the
        function will still succeed if given.
    dist : float
        Distance to project points to in meters.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        Width and height (w, h) of the area intersected by the given frustum at
        `dist`.

    Examples
    --------
    Compute the viewport required to draw in the area where the frustum
    intersects the screen::

        # needed information
        scrWidthM = 0.52
        scrDistM = 0.72
        scrWidthPIX = 1920
        scrHeightPIX = 1080
        scrAspect = scrWidthPIX / float(scrHeightPIX)
        pixPerMeter = scrWidthPIX / scrWidthM

        # Compute a frustum for 20 degree vertical FOV at distance of the
        # screen.
        frustum = computeFrustumFOV(20., scrAspect, scrDistM)

        # get the dimensions of the frustum
        w, h = projectFrustum(frustum, scrDistM) * pixPerMeter

        # get the origin of the viewport, relative to center of screen.
        x = (scrWidthPIX - w) / 2.
        y = (scrHeightPIX - h) / 2.

        # if there is an eye offset ...
        # x = (scrWidthPIX - w + eyeOffsetM * pixPerMeter) / 2.

        # viewport rectangle
        rect = np.asarray((x, y, w, h), dtype=int)

    You can then set the viewport/scissor rectangle of the buffer to restrict
    drawing to `rect`.

    """
    dtype = np.float64 if dtype is None else np.dtype(dtype).type

    frustum = np.asarray(frustum, dtype=dtype)
    l, r, t, b = np.abs(frustum[:4] * dist / frustum[4], dtype=dtype)

    return np.array((l + r, t + b), dtype=dtype)


def projectFrustumToPlane(frustum, planeOrig, dtype=None):
    """Project a frustum on a fronto-parallel plane and get the coordinates of
    the corners in physical space.

    Parameters
    ----------
    frustum : array_like
        Frustum parameters (left, right, bottom, top, near, far), you can
        exclude `far` since it is not used in this calculation. However, the
        function will still succeed if given.
    planeOrig : float
        Distance of plane to project points on in meters.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        4x3 array of coordinates in the physical reference frame with origin
        at the eye.

    """
    dtype = np.float64 if dtype is None else np.dtype(dtype).type

    frustum = np.asarray(frustum, dtype=dtype)
    l, r, t, b = frustum[:4] * planeOrig / frustum[4]
    d = -planeOrig

    return np.array(((l, t, d), (l, b, d), (r, b, d), (r, t, d)), dtype=dtype)


def generalizedPerspectiveProjection(posBottomLeft,
                                     posBottomRight,
                                     posTopLeft,
                                     eyePos,
                                     nearClip=0.01,
                                     farClip=100.0,
                                     dtype=None):
    """Generalized derivation of projection and view matrices based on the
    physical configuration of the display system.

    This implementation is based on Robert Kooima's 'Generalized Perspective
    Projection' method [1]_.

    Parameters
    ----------
    posBottomLeft : list of float or ndarray
        Bottom-left 3D coordinate of the screen in meters.
    posBottomRight : list of float or ndarray
        Bottom-right 3D coordinate of the screen in meters.
    posTopLeft : list of float or ndarray
        Top-left 3D coordinate of the screen in meters.
    eyePos : list of float or ndarray
        Coordinate of the eye in meters.
    nearClip : float
        Near clipping plane distance from viewer in meters.
    farClip : float
        Far clipping plane distance from viewer in meters.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    tuple
        The 4x4 projection and view matrix.

    See Also
    --------
    computeFrustum : Compute frustum parameters.

    Notes
    -----
    * The resulting projection frustums are off-axis relative to the center of
      the display.
    * The returned matrices are row-major. Values are floats with 32-bits
      of precision stored as a contiguous (C-order) array.

    References
    ----------
    .. [1] Kooima, R. (2009). Generalized perspective projection. J. Sch.
       Electron. Eng. Comput. Sci.

    Examples
    --------
    Computing a projection and view matrices for a window::

        projMatrix, viewMatrix = viewtools.generalizedPerspectiveProjection(
            posBottomLeft, posBottomRight, posTopLeft, eyePos)
        # set the window matrices
        win.projectionMatrix = projMatrix
        win.viewMatrix = viewMatrix
        # before rendering
        win.applyEyeTransform()

    Stereo-pair rendering example from Kooima (2009)::

        # configuration of screen and eyes
        posBottomLeft = [-1.5, -0.75, -18.0]
        posBottomRight = [1.5, -0.75, -18.0]
        posTopLeft = [-1.5, 0.75, -18.0]
        posLeftEye = [-1.25, 0.0, 0.0]
        posRightEye = [1.25, 0.0, 0.0]
        # create projection and view matrices
        leftProjMatrix, leftViewMatrix = generalizedPerspectiveProjection(
            posBottomLeft, posBottomRight, posTopLeft, posLeftEye)
        rightProjMatrix, rightViewMatrix = generalizedPerspectiveProjection(
            posBottomLeft, posBottomRight, posTopLeft, posRightEye)

    """
    # get data type of arrays
    dtype = np.float64 if dtype is None else np.dtype(dtype).type

    # convert everything to numpy arrays
    posBottomLeft = np.asarray(posBottomLeft, dtype=dtype)
    posBottomRight = np.asarray(posBottomRight, dtype=dtype)
    posTopLeft = np.asarray(posTopLeft, dtype=dtype)
    eyePos = np.asarray(eyePos, dtype=dtype)

    # orthonormal basis of the screen plane
    vr = posBottomRight - posBottomLeft
    vr /= np.linalg.norm(vr)
    vu = posTopLeft - posBottomLeft
    vu /= np.linalg.norm(vu)
    vn = np.cross(vr, vu)
    vn /= np.linalg.norm(vn)

    # screen corner vectors
    va = posBottomLeft - eyePos
    vb = posBottomRight - eyePos
    vc = posTopLeft - eyePos

    dist = -np.dot(va, vn)
    nearOverDist = nearClip / dist
    left = np.dot(vr, va) * nearOverDist
    right = np.dot(vr, vb) * nearOverDist
    bottom = np.dot(vu, va) * nearOverDist
    top = np.dot(vu, vc) * nearOverDist

    # projection matrix to return
    projMat = perspectiveProjectionMatrix(
        left, right, bottom, top, nearClip, farClip, dtype=dtype)

    # view matrix to return, first compute the rotation component
    rotMat = np.zeros((4, 4), dtype=dtype)
    rotMat[0, :3] = vr
    rotMat[1, :3] = vu
    rotMat[2, :3] = vn
    rotMat[3, 3] = 1.0

    transMat = np.identity(4, dtype=dtype)
    transMat[:3, 3] = -eyePos

    return projMat, np.matmul(rotMat, transMat)


def orthoProjectionMatrix(left, right, bottom, top, nearClip=0.01, farClip=100.,
                          out=None, dtype=None):
    """Compute an orthographic projection matrix with provided frustum
    parameters.

    Parameters
    ----------
    left : float
        Left clipping plane coordinate.
    right : float
        Right clipping plane coordinate.
    bottom : float
        Bottom clipping plane coordinate.
    top : float
        Top clipping plane coordinate.
    nearClip : float
        Near clipping plane distance from viewer.
    farClip : float
        Far clipping plane distance from viewer.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        4x4 projection matrix

    See Also
    --------
    perspectiveProjectionMatrix : Compute a perspective projection matrix.

    Notes
    -----

    * The returned matrix is row-major. Values are floats with 32-bits of
      precision stored as a contiguous (C-order) array.

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    projMat = np.zeros((4, 4,), dtype=dtype) if out is None else out
    if out is not None:
        projMat.fill(0.0)

    u = dtype(2.0)
    projMat[0, 0] = u / (right - left)
    projMat[1, 1] = u / (top - bottom)
    projMat[2, 2] = -u / (farClip - nearClip)
    projMat[0, 3] = -((right + left) / (right - left))
    projMat[1, 3] = -((top + bottom) / (top - bottom))
    projMat[2, 3] = -((farClip + nearClip) / (farClip - nearClip))
    projMat[3, 3] = 1.0

    return projMat


def perspectiveProjectionMatrix(left, right, bottom, top, nearClip=0.01,
                                farClip=100., out=None, dtype=None):
    """Compute an perspective projection matrix with provided frustum
    parameters. The frustum can be asymmetric.

    Parameters
    ----------
    left : float
        Left clipping plane coordinate.
    right : float
        Right clipping plane coordinate.
    bottom : float
        Bottom clipping plane coordinate.
    top : float
        Top clipping plane coordinate.
    nearClip : float
        Near clipping plane distance from viewer.
    farClip : float
        Far clipping plane distance from viewer.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        4x4 projection matrix

    See Also
    --------
    orthoProjectionMatrix : Compute a orthographic projection matrix.

    Notes
    -----

    * The returned matrix is row-major. Values are floats with 32-bits of
      precision stored as a contiguous (C-order) array.

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    projMat = np.zeros((4, 4,), dtype=dtype) if out is None else out
    if out is not None:
        projMat.fill(0.0)

    u = dtype(2.0)
    projMat[0, 0] = (u * nearClip) / (right - left)
    projMat[1, 1] = (u * nearClip) / (top - bottom)
    projMat[0, 2] = (right + left) / (right - left)
    projMat[1, 2] = (top + bottom) / (top - bottom)
    projMat[2, 2] = -(farClip + nearClip) / (farClip - nearClip)
    projMat[3, 2] = -1.0
    projMat[2, 3] = -(u * farClip * nearClip) / (farClip - nearClip)

    return projMat


def lookAt(eyePos, centerPos, upVec=(0.0, 1.0, 0.0), out=None, dtype=None):
    """Create a transformation matrix to orient a view towards some point. Based
    on the same algorithm as 'gluLookAt'. This does not generate a projection
    matrix, but rather the matrix to transform the observer's view in the scene.

    For more information see:
    https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/gluLookAt.xml

    Parameters
    ----------
    eyePos : list of float or ndarray
        Eye position in the scene.
    centerPos : list of float or ndarray
        Position of the object center in the scene.
    upVec : list of float or ndarray, optional
        Vector defining the up vector. Default is +Y is up.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        4x4 view matrix

    Notes
    -----

    * The returned matrix is row-major. Values are floats with 32-bits of
      precision stored as a contiguous (C-order) array.

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    toReturn = np.zeros((4, 4,), dtype=dtype) if out is None else out
    if out is not None:
        toReturn.fill(0.0)

    eyePos = np.asarray(eyePos, dtype=dtype)
    centerPos = np.asarray(centerPos, dtype=dtype)
    upVec = np.asarray(upVec, dtype=dtype)

    f = centerPos - eyePos
    f /= np.linalg.norm(f)
    upVec /= np.linalg.norm(upVec)

    s = np.cross(f, upVec)
    u = np.cross(s / np.linalg.norm(s), f)

    rotMat = np.zeros((4, 4), dtype=dtype)
    rotMat[0, :3] = s
    rotMat[1, :3] = u
    rotMat[2, :3] = -f
    rotMat[3, 3] = 1.0

    transMat = np.identity(4, dtype=dtype)
    transMat[:3, 3] = -eyePos

    return np.matmul(rotMat, transMat, out=toReturn)


def viewMatrix(pos, ori=(0., 0., 0., -1.), out=None, dtype=None):
    """Get a view matrix from a pose.

    A pose consists of a position coordinate [X, Y, Z, 1] and orientation
    quaternion [X, Y, Z, W]. Assumes that the identity pose has a forward vector
    pointing along the -Z axis and up vector along the +Y axis. The quaternion
    for `ori` must be normalized.

    Parameters
    ----------
    pos : ndarray, tuple, or list of float
        Position vector [x, y, z].
    ori : tuple, list or ndarray of float
        Orientation quaternion in form [x, y, z, w] where w is real and x, y, z
        are imaginary components.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for computations can either be 'float32' or 'float64'. If
        `out` is specified, the data type of `out` is used and this argument is
        ignored. If `out` is not provided, 'float64' is used by default.

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(dtype).type

    # convert if needed
    pos = np.asarray(pos, dtype=dtype)
    ori = np.asarray(ori, dtype=dtype)
    axes = np.asarray(VEC_FWD_AND_UP, dtype=dtype)  # convert to type
    toReturn = np.zeros((4, 4), dtype=dtype) if out is None else out

    # generate rotation matrix
    b, c, d, a = ori[:]
    vsqr = np.square(ori)
    R = np.zeros((3, 3,), dtype=dtype)
    u = dtype(2.0)
    R[0, 0] = vsqr[3] + vsqr[0] - vsqr[1] - vsqr[2]
    R[1, 0] = u * (b * c + a * d)
    R[2, 0] = u * (b * d - a * c)
    R[0, 1] = u * (b * c - a * d)
    R[1, 1] = vsqr[3] - vsqr[0] + vsqr[1] - vsqr[2]
    R[2, 1] = u * (c * d + a * b)
    R[0, 2] = u * (b * d + a * c)
    R[1, 2] = u * (c * d - a * b)
    R[2, 2] = vsqr[3] - vsqr[0] - vsqr[1] + vsqr[2]

    # transform the axes
    transformedAxes = axes.dot(R.T)
    fwdVec = transformedAxes[0, :] + pos
    upVec = transformedAxes[1, :]

    toReturn[:, :] = lookAt(pos, fwdVec, upVec, dtype=dtype)

    return toReturn


def pointToNdc(wcsPos, viewMatrix, projectionMatrix, out=None, dtype=None):
    """Map the position of a point in world space to normalized device
    coordinates/space.

    Parameters
    ----------
    wcsPos : tuple, list or ndarray
        Nx3 position vector(s) (xyz) in world space coordinates.
    viewMatrix : ndarray
        4x4 view matrix.
    projectionMatrix : ndarray
        4x4 projection matrix.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        3x1 vector of normalized device coordinates with type 'float32'

    Notes
    -----
    * The point is not visible, falling outside of the viewing frustum, if the
      returned coordinates fall outside of -1 and 1 along any dimension.
    * In the rare instance the point falls directly on the eye in world
      space where the frustum converges to a point (singularity), the divisor
      will be zero during perspective division. To avoid this, the divisor is
      'bumped' to 1e-5.
    * This function assumes the display area is rectilinear. Any distortion or
      warping applied in normalized device or viewport space is not considered.

    Examples
    --------
    Determine if a point is visible::

        point = (0.0, 0.0, 10.0)  # behind the observer
        ndc = pointToNdc(point, win.viewMatrix, win.projectionMatrix)
        isVisible = not np.any((ndc > 1.0) | (ndc < -1.0))

    Convert NDC to viewport (or pixel) coordinates::

        scrRes = (1920, 1200)
        point = (0.0, 0.0, -5.0)  # forward -5.0 from eye
        x, y, z = pointToNdc(point, win.viewMatrix, win.projectionMatrix)
        pixelX = ((x + 1.0) / 2.0) * scrRes[0])
        pixelY = ((y + 1.0) / 2.0) * scrRes[1])
        # object at point will appear at (pixelX, pixelY)

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    wcsPos = np.asarray(wcsPos, dtype=dtype)  # convert to array
    toReturn = np.zeros_like(wcsPos, dtype=dtype) if out is None else out

    # forward transform from world to clipping space
    viewProjMatrix = np.zeros((4, 4,), dtype=dtype)
    np.matmul(projectionMatrix, viewMatrix, viewProjMatrix)

    pnts, rtn = np.atleast_2d(wcsPos, toReturn)

    # convert to 4-vector with W=1.0
    wcsVec = np.zeros((pnts.shape[0], 4), dtype=dtype)
    wcsVec[:, :3] = wcsPos
    wcsVec[:, 3] = 1.0

    # convert to homogeneous clip space
    wcsVec = mt.applyMatrix(viewProjMatrix, wcsVec, dtype=dtype)

    # handle the singularity where perspective division will fail
    wcsVec[np.abs(wcsVec[:, 3]) <= np.finfo(dtype).eps] = np.finfo(dtype).eps
    rtn[:, :] = wcsVec[:, :3] / wcsVec[:, 3:]  # xyz / w

    return toReturn


def cursorToRay(cursorX, cursorY, winSize, viewport, projectionMatrix,
                normalize=True, out=None, dtype=None):
    """Convert a 2D mouse coordinate to a 3D ray.

    Takes a 2D window/mouse coordinate and transforms it to a 3D direction
    vector from the viewpoint in eye space (vector origin is [0, 0, 0]). The
    center of the screen projects to vector [0, 0, -1].

    Parameters
    ----------
    cursorX, cursorY :  float or int
        Window coordinates. These need to be scaled if you are using a
        framebuffer that does not have 1:1 pixel mapping (i.e. retina display).
    winSize : array_like
        Size of the window client area [w, h].
    viewport : array_like
        Viewport rectangle [x, y, w, h] being used.
    projectionMatrix : ndarray
        4x4 projection matrix being used.
    normalize : bool
        Normalize the resulting vector.
    out : ndarray, optional
        Optional output array. Must be same `shape` and `dtype` as the expected
        output if `out` was not specified.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray
        Direction vector (x, y, z).

    Examples
    --------
    Place a 3D stim at the mouse location 5.0 scene units (meters) away::

        # define camera
        camera = RigidBodyPose((-3.0, 5.0, 3.5))
        camera.alignTo((0, 0, 0))

        # in the render loop

        dist = 5.0
        mouseRay = vt.cursorToRay(x, y, win.size, win.viewport, win.projectionMatrix)
        mouseRay *= dist  # scale the vector

        # set the sphere position by transforming vector to world space
        sphere.thePose.pos = camera.transform(mouseRay)

    """
    if out is None:
        dtype = np.float64 if dtype is None else np.dtype(dtype).type
    else:
        dtype = np.dtype(out.dtype).type

    toReturn = np.zeros((3,), dtype=dtype) if out is None else out

    projectionMatrix = np.asarray(projectionMatrix, dtype=dtype)

    # compute the inverse model/view and projection matrix
    invPM = np.linalg.inv(projectionMatrix)

    # transform psychopy mouse coordinates to viewport coordinates
    cursorX = cursorX + (winSize[0] / 2.0)
    cursorY = cursorY + (winSize[1] / 2.0)

    # get the NDC coordinates of the
    projX = 2. * (cursorX - viewport[0]) / viewport[2] - 1.0
    projY = 2. * (cursorY - viewport[1]) / viewport[3] - 1.0

    vecNear = np.array((projX, projY, 0.0, 1.0), dtype=dtype)
    vecFar = np.array((projX, projY, 1.0, 1.0), dtype=dtype)

    vecNear[:] = vecNear.dot(invPM.T)
    vecFar[:] = vecFar.dot(invPM.T)

    vecNear /= vecNear[3]
    vecFar /= vecFar[3]

    # direction vector
    toReturn[:] = (vecFar - vecNear)[:3]

    if normalize:
        mt.normalize(toReturn, out=toReturn)

    return toReturn


def visibleBBox(extents, mvp, dtype=None):
    """Check if a bounding box is visible.

    This function checks if a bonding box intersects a frustum defined by the
    current projection matrix, after being transformed by the model-view matrix.

    Parameters
    ----------
    extents : array_like
        Bounding box minimum and maximum extents as a 2x3 array. The first row
        if the minimum extents along each axis, and the second row the maximum
        extents (eg. [[minX, minY, minZ], [maxX, maxY, maxZ]]).
    mvp : array_like
        4x4 MVP matrix.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    ndarray or bool
        Visibility test results.

    """
    dtype = np.float64 if dtype is None else np.dtype(dtype).type

    # convert input if needed
    extents = np.asarray(extents, dtype=dtype)
    if not extents.shape == (2, 3):
        raise ValueError("Invalid array dimensions for `extents`.")

    # ensure matrix is array
    mvp = np.asarray(mvp, dtype=dtype)

    # convert BBox to corners
    corners = mt.computeBBoxCorners(extents, dtype=dtype)

    # apply the matrix
    corners = corners.dot(mvp.T)
    # break up into components
    x, y, z = corners[:, 0], corners[:, 1], corners[:, 2]
    wpos, wneg = corners[:, 3], -corners[:, 3]

    # test if box falls all to one side of the frustum
    if np.logical_xor(np.all(x <= wneg), np.all(x >= wpos)):  # x-axis
        return False
    elif np.logical_xor(np.all(y <= wneg), np.all(y >= wpos)):  # y-axis
        return False
    elif np.logical_xor(np.all(z <= wneg), np.all(z >= wpos)):  # z-axis
        return False
    else:
        return True


def visible(points, mvp, mode='discrete', dtype=None):
    """Test if points are visible.

    This function is useful for visibility culling, where objects are only drawn
    if a portion of them are visible. This test can avoid costly drawing calls
    and OpenGL state changes if the object is not visible.

    Parameters
    ----------
    points : array_like
        Point(s) or bounding box to test. Input array must be Nx3 or Nx4, where
        each row is a point. It is recommended that the input be Nx4 since the
        `w` component will be appended if the input is Nx3 which adds overhead.
    mvp : array_like
        4x4 MVP matrix.
    mode : str
        Test mode. If `'discrete'`, rows of `points` are treated as individual
        points. This function will return an array of boolean values with length
        equal to the number of rows in `points`, where the value at each index
        corresponds to the visibility test results for points at the matching
        row index of `points`. If `'group'` a single boolean value is returned,
        which is `False` if all points fall to one side of the frustum.
    dtype : dtype or str, optional
        Data type for arrays, can either be 'float32' or 'float64'. If `None` is
        specified, the data type is inferred by `out`. If `out` is not provided,
        the default is 'float64'.

    Returns
    -------
    bool or ndarray
        Test results. The type returned depends on `mode`.

    Examples
    --------
    Visibility culling, only a draw line connecting two points if visible::

        linePoints = [[-1.0, -1.0, -1.0, 1.0],
                      [ 1.0,  1.0,  1.0, 1.0]]

        mvp = np.matmul(win.projectionMatrix, win.viewMatrix)
        if visible(linePoints, mvp, mode='group'):
            # drawing commands here ...

    """
    dtype = np.float64 if dtype is None else np.dtype(dtype).type

    # convert input if needed
    points = np.asarray(points, dtype=dtype)
    # keep track of dimension, return only a single value if ndim==1
    ndim = points.ndim

    # ensure matrix is array
    mvp = np.asarray(mvp, dtype=dtype)

    # convert to 2d view
    points = np.atleast_2d(np.asarray(points, dtype=dtype))
    if points.shape[1] == 3:  # make sure we are using Nx4
        temp = np.zeros((points.shape[0], 4), dtype=dtype)
        temp[:, :3] = points
        temp[:, 3] = 1.0
        points = temp

    # apply the matrix
    points = points.dot(mvp.T)
    # break up into components
    x, y, z = points[:, 0], points[:, 1], points[:, 2]
    wpos, wneg = points[:, 3], -points[:, 3]

    # test using the appropriate mode
    if mode == 'discrete':
        toReturn = np.logical_and.reduce(
            (x > wneg, x < wpos, y > wneg, y < wpos, z > wneg, z < wpos))
        return toReturn[0] if ndim == 1 else toReturn
    elif mode == 'group':
        # Check conditions for each axis. If all points fall to one side or
        # another, the bounding box is not visible. If all points fall outside
        # of both sides of the frustum along the same axis, that means the box
        # passes through the frustum or the viewer is inside the bounding box
        # and therefore is visible. We do an XOR to capture conditions where all
        # points fall all to one side only. Lastly, if any point is in the
        # bounding box, it will indicate that it's visible.
        #
        # mdc - This has been vectorized to be super fast, however maybe someone
        # smarter than me can figure out something better.
        #
        if np.logical_xor(np.all(x <= wneg), np.all(x >= wpos)):  # x-axis
            return False
        elif np.logical_xor(np.all(y <= wneg), np.all(y >= wpos)):  # y-axis
            return False
        elif np.logical_xor(np.all(z <= wneg), np.all(z >= wpos)):  # z-axis
            return False
        else:
            return True
    else:
        raise ValueError(
            "Invalid `mode` specified, should be either 'discrete' or 'group'.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of basic mouse handling from the ioHub (a separate asynchronous process for
fetching and processing events from hardware; mice, keyboards, eyetrackers).

Initial Version: May 6th, 2013, Sol Simpson
Abbrieviated: May 2013, Jon Peirce
Updated July, 2013, Sol, Added timeouts
"""

from __future__ import absolute_import, division, print_function

import sys

from psychopy import visual, core
from psychopy.iohub import launchHubServer

# create the process that will run in the background polling devices
io = launchHubServer()

# some default devices have been created that can now be used
display = io.devices.display
keyboard = io.devices.keyboard
mouse = io.devices.mouse

# We can use display to find info for the Window creation, like the resolution
# (which means PsychoPy won't warn you that the fullscreen does not match your requested size)
display_resolution = display.getPixelResolution()

# ioHub currently supports the use of a single full-screen PsychoPy Window
win = visual.Window(display_resolution,
                    units='pix', fullscr=True, allowGUI=False, screen=0)

# Create some psychopy visual stim (same as how you would do so normally):
fixSpot = visual.GratingStim(win, tex="none", mask="gauss",
    pos=(0, 0), size=(30, 30), color='black', autoLog=False)
grating = visual.GratingStim(win, pos=(300, 0),
    tex="sin", mask="gauss",
    color=[1.0, 0.5, -1.0],
    size=(150.0, 150.0), sf=(0.01, 0.0),
    autoLog=False)
message = visual.TextStim(win, pos=(0.0, -(display_resolution[1]/3)),
    height=40,
    alignText='center', anchorHoriz='center', anchorVert='center',
    text='move=mv-spot, left-drag=SF, right-drag=mv-grating, scroll=ori',
    autoLog=False, wrapWidth=display_resolution[0] * .9)
message2 = visual.TextStim(win, pos=(0.0, -(display_resolution[1]/4)),
    alignText='center', anchorHoriz='center', anchorVert='center', height=40,
    text='Press Any Key to Quit.',
    autoLog=False, wrapWidth=display_resolution[0] * .9)

last_wheelPosY = 0

io.clearEvents('all')

demo_timeout_start = core.getTime()
# Run the example until a keyboard event is received.

kb_events = None
while not kb_events:
    # Get the current mouse position
    # posDelta is the change in position * since the last call *
    position, posDelta = mouse.getPositionAndDelta()
    mouse_dX, mouse_dY = posDelta

    # Get the current state of each of the Mouse Buttons
    left_button, middle_button, right_button = mouse.getCurrentButtonStates()

    # If the left button is pressed, change the grating's spatial frequency
    if left_button:
        grating.setSF(mouse_dX / 5000.0, '+')
    elif right_button:
        grating.setPos(position)

    # If no buttons are pressed on the Mouse, move the position of the mouse cursor.
    if True not in (left_button, middle_button, right_button):
        fixSpot.setPos(position)

    if sys.platform == 'darwin':
        # On macOS, both x and y mouse wheel events can be detected, assuming the mouse being used
        # supported 2D mouse wheel motion.
        wheelPosX, wheelPosY = mouse.getScroll()
    else:
        # On Windows and Linux, only vertical (Y) wheel position is supported.
        wheelPosY = mouse.getScroll()

    # keep track of the wheel position 'delta' since the last frame.
    wheel_dY = wheelPosY - last_wheelPosY
    last_wheelPosY = wheelPosY

    # Change the orientation of the visual grating based on any vertical mouse wheel movement.
    grating.setOri(wheel_dY * 5, '+')

    # Advance 0.05 cycles per frame.
    grating.setPhase(0.05, '+')

    # Redraw the stim for this frame.
    fixSpot.draw()
    grating.draw()
    message.draw()
    message2.draw()
    flip_time = win.flip()  # redraw the buffer

    # Check for keyboard orand mouse events.
    # If 15 seconds passes without receiving any kb or mouse event,
    # then exit the demo
    kb_events = keyboard.getEvents()
    mouse_events = mouse.getEvents()
    if mouse_events:
        demo_timeout_start = mouse_events[-1].time

    if flip_time - demo_timeout_start > 15.0:
        print("Ending Demo Due to 15 Seconds of Inactivity.")
        break

    # Clear out events that were not accessed this frame.
    io.clearEvents()

# End of Example

win.close()
core.quit()

# The contents of this file are in the public domain.



# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
"""ioHub Common Eye Tracker Interface for Tobii (C) Eye Tracking System"""

from __future__ import absolute_import

from .eyetracker import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

from .calibTools import *
"""# probably only need something like:
    DACrange, GammaCalculator, Monitor, Photometer,
    findPR650, gammaFun, gammaInvFun, getAllMonitors,
    getLumSeries, getLumSeriesPR650, getRGBspectra,
    makeDKL2RGB, makeLMS2RGB,
    monitorFolder, pr650code,
    wavelength_5nm, juddVosXYZ1976_5nm, cones_SmithPokorny
    )
"""

# create a test monitor if there isn't one already
if 'testMonitor' not in getAllMonitors():
    defMon = Monitor('testMonitor',
                     width=30,
                     distance=57,
                     gamma=1.0,
                     # can't always localize the notes easily;
                     # need app created first to import localization and
                     # use _translate( ) => issues
                     notes='default (not very useful) monitor')
    defMon.setSizePix([1024, 768])
    defMon.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Allen Institute for Brain Science

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License Version 3
as published by the Free Software Foundation on 29 June 2007.
This program is distributed WITHOUT WARRANTY OF MERCHANTABILITY OR FITNESS
FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESSED OR IMPLIED.
See the GNU General Public License Version 3 for more details.
You should have received a copy of the GNU General Public License along
with this program. If not, see http://www.gnu.org/licenses/
"""
from __future__ import absolute_import, division, print_function

from builtins import object
from past.utils import old_div
import pyglet
GL = pyglet.gl


class ProjectorFramePacker(object):
    """Class which packs 3 monochrome images per RGB frame.

    Allowing 180Hz stimuli with DLP projectors such as TI LightCrafter 4500.

    The class overrides methods of the visual.Window class to pack a
    monochrome image into each RGB channel. PsychoPy is running at 180Hz.
    The display device is running at 60Hz.  The output projector is producing
    images at 180Hz.

    Frame packing can work with any projector which can operate in
    'structured light mode' where each RGB channel is presented
    sequentially as a monochrome image.  Most home and office projectors
    cannot operate in this mode, but projectors designed for machine
    vision applications typically will offer this feature.

    Example usage to use ProjectorFramePacker::

        from psychopy.visual.windowframepack import ProjectorFramePacker
        win = Window(monitor='testMonitor', screen=1,
                     fullscr=True, useFBO = True)
        framePacker = ProjectorFramePacker (win)
    """

    def __init__(self, win):
        """
        :Parameters:

            win : Handle to the window.

        """
        super(ProjectorFramePacker, self).__init__()
        self.win = win
        # monkey patch window
        win._startOfFlip = self.startOfFlip
        win._endOfFlip = self.endOfFlip

        # This part is increasingly ugly.  Add a function to set these values?
        win._monitorFrameRate = 180.0
        win.monitorFramePeriod = old_div(1.0, win._monitorFrameRate)
        win.refreshThreshold = (old_div(1.0, win._monitorFrameRate)) * 1.2

        # enable Blue initially, since projector output sequence is BGR
        GL.glColorMask(False, False, True, True)
        self.flipCounter = 0

    def startOfFlip(self):
        """Return True if all channels of the RGB frame have been filled
        with monochrome images, and the associated window should perform
        a hardware flip"""
        return self.flipCounter % 3 == 2

    def endOfFlip(self, clearBuffer):
        """Mask RGB cyclically after each flip.
        We ignore clearBuffer and just auto-clear after each hardware flip.
        """
        if self.flipCounter % 3 == 2:
            GL.glClear(GL.GL_COLOR_BUFFER_BIT)

        self.flipCounter += 1
        if self.flipCounter % 3 == 0:
            GL.glColorMask(False, True, False, True)  # enable green
        elif self.flipCounter % 3 == 1:
            GL.glColorMask(True, False, False, True)  # enable red
        elif self.flipCounter % 3 == 2:
            GL.glColorMask(False, False, True, True)  # enable blue



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

import pytest

from psychopy.visual.window import Window
from psychopy.visual.slider import Slider
from psychopy.visual.grating import GratingStim
from psychopy.visual.elementarray import ElementArrayStim
from psychopy.visual.circle import Circle
from psychopy.visual.rect import Rect
from psychopy import constants
from numpy import array_equal
import random


class Test_Slider(object):
    def setup_class(self):
        self.win = Window([128,128], pos=[50,50], allowGUI=False,
                          autoLog=False)

    def teardown_class(self):
        self.win.close()

    def test_color(self):
        colors = ['black', 'red']

        for color in colors:
            s = Slider(self.win, color=color)

            assert s.line.color == color
            assert s.tickLines.colors == color

            for l in s.labelObjs:
                assert l.color == color

    def test_change_color(self):
        s = Slider(self.win, color='black')

        with pytest.raises(AttributeError):
            s.color = 'blue'

    def test_size(self):
        sizes = [(1, 0.1), (1.5, 0.5)]

        for size in sizes:
            s = Slider(self.win, size=size)
            assert s.size == size

    def test_change_size(self):
        s = Slider(self.win, size=(1, 0.1))

        with pytest.raises(AttributeError):
            s.size = (1.5, 0.5)

    def test_lineLength(self):
        s = Slider(self.win, size=(1, 0.1))
        assert s._lineL == 1

    def test_tickWidth(self):
        s = Slider(self.win, size=(1, 0.1))
        assert s._lineW == (1 * s._lineAspectRatio)

    def test_horiz(self):
        s = Slider(self.win, size=(1, 0.1))
        assert s.horiz == True
        s = Slider(self.win, size=(0.1, 1))
        assert s.horiz == False

    def test_reset(self):
        s = Slider(self.win, size=(1, 0.1))
        s.markerPos = 1
        s.history = [1]
        s.rating = 1
        s.rt = 1
        s.status = constants.STARTED
        s.reset()
        assert s.markerPos == None
        assert s.history == []
        assert s.rating == None
        assert s.rt == None
        assert s.status == constants.NOT_STARTED

    def test_elements(self):
        s = Slider(self.win, size=(1, 0.1))
        assert type(s.line) == type(GratingStim(self.win))
        assert type(s.tickLines) == type(ElementArrayStim(self.win))
        assert type(s.marker) == type(Circle(self.win))
        assert type(s.validArea) == type(Rect(self.win))
        
    def test_pos(self):
        s = Slider(self.win, size=(1, 0.1))
        positions = [(.05, .05),(0.2, .2)]
        for newPos in positions:
            s.pos = newPos
            assert array_equal(s.pos, newPos)
        
    def test_ratingToPos(self):
        s = Slider(self.win, size=(1, 0.1), )
        assert s._ratingToPos(3)[0][0] == 0
        assert s._ratingToPos(1)[0][0] == -.5
        assert s._ratingToPos(5)[0][0] == .5

    def test_posToRatingToPos(self):
        s = Slider(self.win, size=(1, 0.1), )
        assert s._posToRating((0, 0)) == 3
        assert s._posToRating((-.5, 0)) == 1
        assert s._posToRating((.5, 0)) == 5

    def test_tickLocs(self):
        s = Slider(self.win, size=(1, 0.1), )
        assert s.tickLocs[0][0] == -.5 and s.tickLocs[0][1] == 0.0
        assert s.tickLocs[1][0] == -.25 and s.tickLocs[1][1] == 0.0
        assert s.tickLocs[2][0] == .0 and s.tickLocs[2][1] == 0.0
        assert s.tickLocs[3][0] == .25 and s.tickLocs[3][1] == 0.0
        assert s.tickLocs[4][0] == .5 and s.tickLocs[4][1] == 0.0

    def test_labelLocs(self):
        s = Slider(self.win, size=(1, 0.1), labels=('a','b','c','d','e'))
        assert s.labelLocs[0][0] == -.5 and s.labelLocs[0][1] == -.1
        assert s.labelLocs[1][0] == -.25 and s.labelLocs[1][1] == -.1
        assert s.labelLocs[2][0] == .0 and s.labelLocs[2][1] == -.1
        assert s.labelLocs[3][0] == .25 and s.labelLocs[3][1] == -.1
        assert s.labelLocs[4][0] == .5 and s.labelLocs[4][1] == -.1

    def test_granularity(self):
        s = Slider(self.win, size=(1, 0.1), granularity=1)
        minRating, maxRating = 1, 5

        assert s.granularity == 1
        assert s._granularRating(1) == minRating
        assert s._granularRating(5) == maxRating
        assert s._granularRating(0) == minRating
        assert s._granularRating(6) == maxRating

    def test_rating(self):
        s = Slider(self.win, size=(1, 0.1))
        minRating, maxRating = 1, 5

        s.rating = 1
        assert s.rating == minRating
        s.rating = 5
        assert s.rating == maxRating
        s.rating = 0
        assert s.rating == minRating
        s.rating = 6
        assert s.rating == maxRating

    def test_markerPos(self):
        s = Slider(self.win, size=(1, 0.1))
        s._updateMarkerPos = False
        minPos, maxPos = 1, 5

        assert s._updateMarkerPos != True
        s.markerPos = 1
        assert s.markerPos == minPos
        s.markerPos = 5
        assert s.markerPos == maxPos
        s.markerPos = 0
        assert s.markerPos == minPos
        s.markerPos = 6
        assert s.markerPos == maxPos
        assert s._updateMarkerPos == True

    def test_recordRating(self):
        s = Slider(self.win, size=(1, 0.1))
        minRating, maxRating = 1, 5

        counter = 0
        for rates in range(0,7):
            s.recordRating(rates, random.random())
            counter +=1

        ratings = [rating[0] for rating in s.history]
        RT = [rt[1] for rt in s.history]

        assert len(s.history) == counter
        assert len(ratings) == counter
        assert min(ratings) == minRating
        assert max(ratings) == maxRating
        assert len(RT) == counter
        assert max(RT) <= 1
        assert min(RT) >= 0

    def test_getRating(self):
        s = Slider(self.win, size=(1, 0.1))
        minRating, maxRating = 1, 5
        s.rating = 1
        assert s.getRating() == minRating
        s.rating = 5
        assert s.getRating() == maxRating
        s.rating = 0
        assert s.getRating() == minRating
        s.rating = 6
        assert s.getRating() == maxRating

    def test_getRT(self):
        s = Slider(self.win, size=(1, 0.1))
        testRT = .89
        s.recordRating(2, testRT)
        assert s.history[-1][1] == s.getRT()
        assert type(s.getRT()) == float
        assert s.getRT() == testRT



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function
from builtins import super  # provides Py3-style super() using python-future

from os import path
from psychopy.experiment.components import BaseVisualComponent, Param, getInitVals, _translate

# the absolute path to the folder containing this path
thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'envelopegrating.png')
tooltip = _translate('Envelope Grating: present cyclic textures including 2nd order envelope stimuli, prebuilt or from a '
                     'file')

# only use _localized values for label values, nothing functional:
_localized = {'carrier': _translate('Carrier texture'),
              'ori': _translate('Carrier Orientation'),
              'mask': _translate('Mask'),
              'sf': _translate('Carrier spatial frequency'),
              'phase': _translate('Carrier phase (in cycles)'),
              'contrast': _translate('Carrier contrast'),
              'texture resolution': _translate('Texture resolution'),
              'interpolate': _translate('Interpolate'),
              'envelope': _translate('Envelope texture'),
              'envsf':_translate('Envelope spatial frequency'),
              'envori':_translate('Envelope orientation'),
              'envphase':_translate('Envelope phase'),
              'moddepth':_translate('Envelope modulation depth'),
              'power':_translate('Power to which envelope is raised'),
              'beat':_translate('Is modulation a beat'),
              'blendmode':_translate('OpenGL blend mode')
              }


class EnvGratingComponent(BaseVisualComponent):
    """A class for presenting grating stimuli"""

    def __init__(self, exp, parentName, name='env_grating', carrier='sin',
                 mask='None', sf=1.0, interpolate='linear',
                 units='from exp settings', color='$[1,1,1]', colorSpace='rgb',
                 pos=(0, 0), size=(0.5, 0.5), ori=0, phase=0.0, texRes='128',
                 envelope='sin',envsf=1.0,envori=0.0,envphase=0.0, 
                 beat=False, power=1.0,
                 contrast=0.5, moddepth=1.0, blendmode='avg',
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim=''):
        super().__init__(
            exp, parentName, name=name, units=units,
            color=color, colorSpace=colorSpace,
            pos=pos, size=size, ori=ori,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        self.type = 'EnvGrating'
        self.url = "http://www.psychopy.org/builder/components/EnvelopeGrating.html"
        self.order = ['carrier', 'mask']

        # params

        self.params['ori'] = Param(
            ori, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=_translate("Orientation of this stimulus (in deg)"),
            label=_localized['ori'],categ="Carrier")

        msg = _translate("The (2D) texture of the background - can be sin, sqr,"
                         " sinXsin... or a filename (including path)")
        self.params['carrier'] = Param(
            carrier, valType='str', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['carrier'], categ="Carrier")

        msg = _translate("An image to define the alpha mask (ie shape)- "
                         "gauss, circle... or a filename (including path)")
        self.params['mask'] = Param(
            mask, valType='str', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['mask'], categ="Carrier")

        msg = _translate("Contrast of background carrier")
        self.params['contrast'] = Param(
            contrast, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['contrast'], categ="Carrier")

        msg = _translate("Spatial frequency of background carrier repeats across the "
                         "grating in 1 or 2 dimensions, e.g. 4 or [2,3]")
        self.params['sf'] = Param(
            sf, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['sf'], categ="Carrier")

        msg = _translate("Spatial positioning of the background carrier "
                         "(wraps in range 0-1.0)")
        self.params['phase'] = Param(
            phase, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['phase'], categ="Carrier")

        msg = _translate(
            "Resolution of the texture for standard ones such as sin, sqr "
            "etc. For most cases a value of 256 pixels will suffice")
        self.params['texture resolution'] = Param(
            texRes,
            valType='code', allowedVals=['32', '64', '128', '256', '512'],
            updates='constant', allowedUpdates=[],
            hint=msg,
            label=_localized['texture resolution'], categ="Carrier")

        msg = _translate("How should the image be interpolated if/when "
                         "rescaled")
        self.params['interpolate'] = Param(
            interpolate, valType='str', allowedVals=['linear', 'nearest'],
            updates='constant', allowedUpdates=[],
            hint=msg,
            label=_localized['interpolate'], categ="Carrier")


        msg = _translate("The (2D) texture of the envelope - can be sin, sqr,"
                         " sinXsin... or a filename (including path)")
        self.params['envelope'] = Param(
            envelope, valType='str', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['envelope'], categ="Envelope")

        msg = _translate("Spatial frequency of the modulation envelope repeats across the "
                         "grating in 1 or 2 dimensions, e.g. 4 or [2,3]")
        self.params['envsf'] = Param(
            envsf, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['envsf'], categ="Envelope")

        msg = _translate("Spatial positioning of the modulation envelope"
                         "(wraps in range 0-1.0)")
        self.params['envphase'] = Param(
            envphase, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['envphase'], categ="Envelope")

        msg = _translate("Orientation of the modulation envelope"
                         "(wraps in range 0-360)")
        self.params['envori'] = Param(
            envori, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['envori'], categ="Envelope")

        msg = _translate("Modulation depth of modulation envelope")
        self.params['moddepth'] = Param(
            moddepth, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['moddepth'], categ="Envelope")
            
        msg = _translate("Power of modulation envelope. "
                          "The modulator will be raised to this power "
                          "according to the equation S=cC*(1+mM)^power, "
                          "where C is the carrier and M is the modulator. "
                          "and c and m are there respective contrast and modulation depth. "
                          "Only works with AM envelopes (hence +1) in "
                          "equation. Power is ignored if a beat is requested. "
                          "This is used to obtain the square root of the modulator (power = 0.5) "
                          "which is useful if combining two envelope gratings "
                          "with different carriers and a 180 degree phase shift "
                          "as the resulting combined signal will not "
                          "have any reduction in local contrast at any point in the image. "
                          "This is similar - but not identical to - the method used by "
                          "Landy and Oruc, Vis Res 2002. "
                          "Note overall contrast (apparent carrier contrast) will be altered.")
        self.params['power'] = Param(
            moddepth, valType='code', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['power'], categ="Envelope")

        msg = _translate("Do you want a 'beat'? [beat = carrier*envelope, "
                         "no beat = carrier*(1+envelope), True/False, Y/N]")
        self.params['beat'] = Param(
            beat, valType='str', allowedTypes=[],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['beat'], categ="Envelope")

        msg = _translate("OpenGL Blendmode. Avg is most common mode"
                         " in PsychoPy, add is useful if combining a beat with"
                         " the carrier image or numpy array at point of display")
        self.params['blendmode'] = Param(
            blendmode, valType='str', allowedVals=['avg', 'add'],
            updates='constant',
            allowedUpdates=['constant', 'set every repeat', 'set every frame'],
            hint=msg,
            label=_localized['blendmode'], categ="Basic")

    def writeInitCode(self, buff):
        # do we need units code?
        if self.params['units'].val == 'from exp settings':
            unitsStr = ""
        else:
            unitsStr = "units=%(units)s, " % self.params
        #buff.writeIndented("from psychopy.visual.secondorder import EnvelopeGrating\n")

        # replaces variable params with defaults and sets sample updating flag
        inits = getInitVals(self.params)

        code = ("%s = visual.EnvelopeGrating(\n" % inits['name'] +
                "    win=win, name='%s',%s\n" % (inits['name'], unitsStr) +
                "    carrier=%(carrier)s, mask=%(mask)s,\n" % inits +
                "    ori=%(ori)s, pos=%(pos)s, size=%(size)s,\n" % inits +
                "    sf=%(sf)s, phase=%(phase)s,\n" % inits +
                "    color=%(color)s, colorSpace=%(colorSpace)s,\n " % inits +
                "    opacity=%(opacity)s, contrast=%(contrast)s,\n" % inits +
                "    texRes=%(texture resolution)s, envelope=%(envelope)s,\n" % inits +
                "    envori=%(envori)s, envsf=%(envsf)s,\n" % inits +
                "    envphase=%(envphase)s, power=%(power)s,\n" % inits +
                "    moddepth=%(moddepth)s, blendmode=%(blendmode)s" %inits )

        if self.params['beat'].val in ['Y','y','Yes', 'yes','True','true']:
            code += ", beat=True"
        elif self.params['beat'].val in ['N','n','No', 'no','False','false']:
            code += ", beat=False"
        else:
            code += ", beat=%(beat)s" %inits

        if self.params['interpolate'].val == 'linear':
            code += ", interpolate=True"
        else:
            code += ", interpolate=False"
        depth = -self.getPosInRoutine()
        code += ", depth=%.1f)\n" % depth
        code += "if sys.version[0]=='3' and np.min(win.gamma) == None:\n"
        code += "    logging.warning('Envelope grating in use with no gamma set. Unless you have hardware gamma correction the image will be distorted.')\n"           
        code += "elif np.min(win.gamma) < 1.01:\n"
        code += "    logging.warning('Envelope grating in use with window gamma <= 1.0 or no gamma set at all. Unless you have hardware gamma correction the image will be distorted.')\n"        
        buff.writeIndentedLines(code)

    def writeRoutineStartCode(self,buff):
        super().writeRoutineStartCode(buff)
        #if self.params['blendmode'].val!='default':
            #buff.writeIndented("__allEnvSaveBlendMode=win.blendMode #required to clean up after %(name)s\n" %self.params)

    def writeFrameCode(self, buff):
        """Write the code that will be called every frame
        """
        buff.writeIndented("\n")
        buff.writeIndented("# *%s* updates\n" % self.params['name'])
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCode(buff)
        buff.writeIndented("%(name)s.setAutoDraw(True)\n" % self.params)
        #if self.params['blendmode'].val!='default':
            #buff.writeIndented("%(name)s_SaveBlendMode=win.blendMode\n" %self.params)
            #buff.writeIndented("win.blendMode=%(blendmode)s\n" %self.params)
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)

        # test for stop (only if there was some setting for duration or stop)
        if self.params['stopVal'].val not in ('', None, -1, 'None'):
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)
            buff.writeIndented("%(name)s.setAutoDraw(False)\n" % self.params)
            #if self.params['blendmode'].val!='default':
                #buff.writeIndented("win.blendMode=%(name)s_SaveBlendMode\n" % self.params)
            # to get out of the if statement
            buff.setIndentLevel(-2, relative=True)

        # set parameters that need updating every frame
        # do any params need updating? (this method inherited from _base)
        if self.checkNeedToUpdate('set every frame'):
            code = "if %(name)s.status == STARTED:  # only update if drawing\n"
            buff.writeIndented(code % self.params)
            buff.setIndentLevel(+1, relative=True)  # to enter the if block
            self.writeParamUpdates(buff, 'set every frame')
            buff.setIndentLevel(-1, relative=True)  # to exit the if block

    def writeRoutineEndCode(self, buff):
        super().writeRoutineEndCode(buff)  # adds start/stop times to data
        #if self.params['blendmode'].val!='default':
            #buff.writeIndented("win.blendMode=__allEnvSaveBlendMode #clean up for %(name)s\n" %self.params)




# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2013 Josh Borah
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
#
# fileauthor:: Sol Simpson <sol@isolver-software.com>
# fileauthor:: Josh Borah <josh@a-s-l.com>

from math import atan, tan, sqrt

#
# distToPixel
#
# Convert between distance coordinates and pixel coordinates.
#
# Distance coordinates are 2D Cartesian coordinates, measured from an origin at the
# center pixel,  and are real distance units (inches, centimeters, etc.) along horizontal and
# vertical screen axes.
#


def distToPixel(
        hpix_per_dist_unit,
        vpix_per_dist_unit,
        pixHres,
        pixVres,
        distH,
        distV):
    pixH = pixHres / 2.0 + (distH * hpix_per_dist_unit)
    pixV = pixVres / 2.0 + (distV * vpix_per_dist_unit)
    return pixH, pixV


def pixelToDist(
        hpix_per_dist_unit,
        vpix_per_dist_unit,
        pixHres,
        pixVres,
        pixH,
        pixV):
    distH = (pixH - pixHres / 2.0) / hpix_per_dist_unit
    distV = (pixV - pixVres / 2.0) / vpix_per_dist_unit
    return distH, distV

#
# All of following assume a nominal eye point 'eye2display' distance units from display
# with line-of-gaze normal to the display at the display center.  Angle variable are
# assumed to have units of degrees.
#
# Since the Python math lib trig functions work with radians,
# a radian to angle conversion factor (deg/rad = 57.2958) is included to give angle
# variables 'degree' units.
#

#
# Convert between distance coordinates (distH, distV) and 'normalized Cartesian
# coordinates' (ndH, ndV).
#
# 'Normalized Cartesian coordinates' are Cartesian distance coordinates, normalized by
# by the distance from the nominal eye point to the display.  For very small distances
# from the origin, these values coorespond to visual angle from the origin along the
# horizontal and vertical display axes. A factor of 57.2958 is used so that the values
# correspond to degrees rather than radians.
#


def convertDistToNd(eye2display, distH, distV):
    ndH = 57.2958 * distH / eye2display
    ndV = 57.2958 * distV / eye2display
    return ndH, ndV


def convertNdToDist(eye2display, ndH, ndV):
    distH = ndH * eye2display / 57.2958
    distV = ndV * eye2display / 57.2958
    return distH, distV

#
# Convert between distance coordinates (distH, distV) and
# 'Cartesian Angles' (caH, caV).
# 'Cartesian Angles' are visual angles (from nominal eye point) along
# horizontal and vertical display axes.  In other words, the horizontal coordinate is the
# visual angle between the origin and the intersection of the Cartesian
# coordinate line with the horizontal axes.
#


def distToCa(eye2display, distH, distV):
    caH = 57.2958 * atan(distH / eye2display)
    caV = 57.2958 * atan(distV / eye2display)
    return caH, caV


def caToDist(eye2display, caH, caV):
    distH = eye2display * tan(caH / 57.2958)
    distV = eye2display * tan(caV / 57.2968)
    return distH, distV


#
# Convert between distance coordinates (distH, distV) and Fick Coordinates (as,el)
#
def distToFick(eye2display, distH, distV):
    az = 57.2958 * atan(distH / eye2display)
    el = 57.2958 * atan(distV / sqrt(eye2display *
                                     eye2display + distH * distH))
    return az, el


def fickToDist(eye2display, az, el):
    distH = eye2display * tan(az / 57.2958)
    distV = sqrt(eye2display * eye2display + distH * distH) * tan(el / 57.2958)
    return distH / distV

#
# Convert between distance coordinates (distH, distV) and 'symmetric angle'
# coordinates (saH, saV).
# 'Symmetric angles' are visual angles between a point on the display and the central
# axes lines, measured along lines parallel to the display axes.  The vertical coordinate is
# same as the Fick elevation angle.  The horizontal coordinate is measured in a
# symmetrical fashion and is not the same as the Fick azimuth angle.
#


def distToSa(eye2display, distH, distV):
    saH = 57.2958 * atan(distH / sqrt(eye2display *
                                      eye2display + distV * distV))
    saV = 57.2958 * atan(distV / sqrt(eye2display *
                                      eye2display + distH * distH))
    return saH, saV


def saToDist(eye2dsply, saH, saV):
    tansaV_sqrd = tan(saV / 57.2958) * tan(saV / 57.2958)
    tansaH_sqrd = tan(saH / 57.2958) * tan(saH / 57.2958)
    Dsqrd = eye2dsply * eye2dsply

    signsaV = 1.0
    if saV < 0.0:
        signsaV = -1.0

    signsaH = 1.0
    if saH < 0.0:
        signsaH = -1.0

    distV = signsaV * sqrt((Dsqrd * tansaV_sqrd + Dsqrd *
                            tansaH_sqrd * tansaV_sqrd) / (1 - tansaH_sqrd * tansaV_sqrd))

    distH = signsaH * sqrt((Dsqrd + distV * distV) * tansaH_sqrd)

    return distV, distH

#-------------------------------------
# Old code using matrix multiplication to convert between screen pix and
# experiment software coord system...

#        if coord_type=='pix':

#        origin=self.getOrigin()
#        if origin not in Display._supported_origin_types:
#            print2err(" *** Display device error: Unknown origin type: {0}".format(origin))
#            return
#
#        x1,y1,x2,y2=self.getBounds()
#        print2err('getBounds: ',self.getBounds(),  )
#
#        bounds_matrix=np.matrix([[x1,y1,1,0],[-y1,x1,0,1],[x2,y2,1,0],[-y2,x2,0,1]])
#
#        cx1=None
#        cy1=None
#        cx2=None
#        cy2=None
#
##        if coord_type == 'org':
##            cx1=x1
##            cy1=y1
##            cx2=x2
##            cy2=y2
#        if coord_type == 'pix':
#            if origin == 'center':
#                cx1=-pixel_width/2.0
#                cy1=pixel_height/2.0
#                cx2=pixel_width/2.0
#                cy2=-pixel_height/2.0
##            elif origin == 'top_left':
##                cx1=0
##                cy1=0
##                cx2=pixel_width
##                cy2=pixel_height
##            elif origin == 'bottom_left':
##                cx1=0
##                cy1=pixel_height
##                cx2=pixel_width
##                cy2=0
#        elif coord_type in ['mm','cm','inch']:
#            phys_to_coord_ratio=1.0
##            if coord_type == 'mm':
##                if phys_unit_type == 'cm':
##                    phys_to_coord_ratio=10.0
##                elif phys_unit_type == 'inch':
##                    phys_to_coord_ratio=25.4
#            if coord_type == 'cm':
#                if phys_unit_type == 'mm':
#                    phys_to_coord_ratio=0.1
#                elif phys_unit_type == 'inch':
#                    phys_to_coord_ratio=2.54
##            elif coord_type == 'inch':
##                if phys_unit_type == 'mm':
##                    phys_to_coord_ratio=0.0393701
##                elif phys_unit_type == 'cm':
##                    phys_to_coord_ratio=0.393701
#
#            if origin == 'center':
#                phys_to_coord_ratio=2.0*phys_to_coord_ratio
#                cx1=-phys_width/phys_to_coord_ratio
#                cy1=phys_height/phys_to_coord_ratio
#                cx2=phys_width/phys_to_coord_ratio
#                cy2=-phys_height/phys_to_coord_ratio
##            elif origin == 'top_left':
##                cx1=0.0
##                cy1=0.0
##                cx2=phys_width*phys_to_coord_ratio
##                cy2=phys_height*phys_to_coord_ratio
##            elif origin == 'bottom_left':
##                cx1=0.0
##                cy1=phys_height*phys_to_coord_ratio
##                cx2=phys_width*phys_to_coord_ratio
##                cy2=0.0
#        elif coord_type == 'norm':
#            if origin == 'center':
#                cx1=-1.0
#                cy1=1.0
#                cx2=1.0
#                cy2=-1.0
##            elif origin == 'top_left':
##                cx1=0.0
##                cy1=0.0
##                cx2=1.0
##                cy2=1.0
##            elif origin == 'bottom_left':
##                cx1=0.0
##                cy1=1.0
##                cx2=1.0
##                cy2=0.0
##        elif coord_type == 'percent':
##            if origin == 'center':
##                cx1=-50.0
##                cy1=50.0
##                cx2=50.0
##                cy2=-50.0
##            elif origin == 'top_left':
##                cx1=0.0
##                cy1=0.0
##                cx2=100.0
##                cy2=100.0
##            elif origin == 'bottom_left':
##                cx1=0.0
##                cy1=100.0
##                cx2=100.0
##                cy2=0.0
#        elif coord_type == 'deg':
#            if origin == 'center':
#                cx1=-degree_width/2.0
#                cy1=degree_height/2.0
#                cx2=degree_width/2.0
#                cy2=-degree_height/2.0
##            elif origin == 'top_left':
##                cx1=0.0
##                cy1=0.0
##                cx2=degree_width
##                cy2=degree_height
##            elif origin == 'bottom_left':
##                cx1=0.0
##                cy1=degree_height
##                cx2=degree_width
##                cy2=0.0
#
#        if cx1 is not None and cy1 is not None  and cx2 is not None and cy2 is not None :
#            coord_matrix=np.matrix( [[cx1],[cy1],[cx2],[cy2]] )
#            abcd = np.linalg.solve(bounds_matrix, coord_matrix)
#            a,b,c,d=np.array(abcd)[:,0]
#            #print2err('abcd: {0}\n a={1}, b={2} , c={3}, d={4}'.format(abcd,a,b,c,d))
#
#
#            def pix2coord(self, x,y,display_index=None):
#                #print2err('Display {0} bounds: {1}'.format(display_index,self.getBounds()))
#                if display_index == self.getIndex():
#                    return a*x+b*y+c, b*x-a*y+d
#                return x,y
#
#            self._pix2coord=pix2coord
#
#            def coord2pix(self,cx,cy,display_index=None):
#                if display_index == self.getIndex():
#                    aabb=(a**2+b**2)
#                    return (a*cx+b*cy-b*d-a*c)/aabb, (b*cx-a*cy-b*c+a*d)/aabb
#                return cx,cy
#
#            self._coord2pix=coord2pix

